from typing import List

from devicemanager.interactor.dto import DeviceStatusDto


class UpdateDevicesStatusResponse(object):
    def __init__(
        self,
        updated,    # type: bool
        devices,    # type: List[DeviceStatusDto]
    ):
        # type: (...) -> None

        self.updated = updated
        self.devices = devices
