class Reports(object):

    canceled_receipt = "CANCELED_RECEIPT"
    canceled_fiscal_coupon = "CANCELED_FISCAL_COUPON"
    end_of_sale = "END_OF_SALE"
    pre_account = "PRE_ACCOUNT"
    conference_items = "CONFERENCE_ITEMS"
    pre_account_by_seats = "PRE_ACCOUNT_BY_SEATS"
    conference_items_by_seats = "CONFERENCE_ITEMS_BY_SEATS"
    pre_account_quick_service = "PRE_ACCOUNT_QUICK_SERVICE"
    kitchen_label = "KITCHEN_LABEL"
