from typing import Dict


class LoyaltyEndpointsConfig(object):
    def __init__(
        self,           # type: LoyaltyEndpointsConfig
        ask_totaled,    # type: bool
        ask_payment,    # type: bool
    ):
        # type: (...) -> None

        self.ask_totaled = ask_totaled
        self.ask_payment = ask_payment
