from datetime import datetime

from ._PublicationAction import PublicationAction


class PublicationHistory(object):
    def __init__(
        self,
        created_at: datetime,
        action: PublicationAction,
    ):
        self.created_at = created_at
        self.action = action
