class EndOfSaleNotesRequest(object):

    def __init__(
        self,       # type: EndOfSaleNotesRequest
        pos_id,     # type: int
        order_id,   # type: int
    ):
        # type: (...) -> None

        self.pos_id = pos_id
        self.order_id = order_id
