from typing import (
    Dict,
    List,
    Union,
)

from nreports.repository import (
    I18NRepository,
    L10NRepository,
)
from nreports_api.shared import JSON
from ._TextLine import TextLine


class IgnoredLine(TextLine):

    def inner_format(
        self,
        data: Dict[str, JSON],
        i18n_repository: I18NRepository,
        l10n_repository: L10NRepository,
    ) -> List[Dict[str, Union[str, Dict[str, str]]]]:

        return [self.line]

    def get_inner_data_to_build(
        self,
    ) -> Dict[str, Union[Dict[str, None], None]]:

        return {}
