class LoyaltyTranslatedMessage(object):
    def __init__(
        self,   # type: LoyaltyTranslatedMessage
        pt,     # type: str
        en,     # type: str
        es,     # type: str
    ):
        # type: (...) -> None

        self.en = en
        self.es = es
        self.pt = pt
